
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include "typeinfo.h"

#define NORTH_SEND_TAG 1000
#define SOUTH_SEND_TAG 2000
#define EAST_SEND_TAG  3000
#define WEST_SEND_TAG  4000
#define BACK_SEND_TAG  5000
#define FRONT_SEND_TAG 6000

SCALAR_TYPE* north_send = NULL;
SCALAR_TYPE* south_send = NULL;
SCALAR_TYPE* north_recv = NULL;
SCALAR_TYPE* south_recv = NULL;
SCALAR_TYPE* east_send = NULL;
SCALAR_TYPE* west_send = NULL;
SCALAR_TYPE* east_recv = NULL;
SCALAR_TYPE* west_recv = NULL;
SCALAR_TYPE* front_send = NULL;
SCALAR_TYPE* back_send = NULL;
SCALAR_TYPE* front_recv = NULL;
SCALAR_TYPE* back_recv = NULL;

int north_size;
int south_size;
int east_size;
int west_size;
int back_size;
int front_size;

void init_buffers(
	const int num_vars, 
	const int nx, 
	const int ny, 
	const int nz,
	const int pe_north,
	const int pe_south,
	const int pe_east,
	const int pe_west,
	const int pe_back,
	const int pe_front) {


	if(pe_north > -1) {
		north_size = sizeof(SCALAR_TYPE) * num_vars * (nx+2) * (nz+2);
	
		north_send = (SCALAR_TYPE*) malloc(north_size);
		north_recv = (SCALAR_TYPE*) malloc(north_size);

	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_south > -1) {
		south_size = sizeof(SCALAR_TYPE) * num_vars * (nx+2) * (nz+2);
	
		south_send = (SCALAR_TYPE*) malloc(south_size);
		south_recv = (SCALAR_TYPE*) malloc(south_size);
	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_east > -1) {
		east_size = sizeof(SCALAR_TYPE) * num_vars * (ny+2) * (nz+2);
	
		east_send = (SCALAR_TYPE*) malloc(east_size);
		east_recv = (SCALAR_TYPE*) malloc(east_size);
	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_west > -1) {
		west_size = sizeof(SCALAR_TYPE) * num_vars * (ny+2) * (nz+2);
	
		west_send = (SCALAR_TYPE*) malloc(west_size);
		west_recv = (SCALAR_TYPE*) malloc(west_size);
	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_front > -1) {
		front_size = sizeof(SCALAR_TYPE) * num_vars * (nx+2) * (ny+2);
	
		front_send = (SCALAR_TYPE*) malloc(front_size);
		front_recv = (SCALAR_TYPE*) malloc(front_size);
	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_back > -1) {
		back_size = sizeof(SCALAR_TYPE) * num_vars * (nx+2) * (ny+2);
	
		back_send = (SCALAR_TYPE*) malloc(back_size);
		back_recv = (SCALAR_TYPE*) malloc(back_size);
	}
}

void exchange(
	const SCALAR_TYPE**** data,
	SCALAR_TYPE**** result,
	const int num_vars, 
	const int nx, 
	const int ny, 
	const int nz,
	const int pe_north,
	const int pe_south,
	const int pe_east,
	const int pe_west,
	const int pe_back,
	const int pe_front) {
	
	int req_counter = 0;
	MPI_Request requests[12];
	
	int nx_counter, ny_counter, nz_counter, nv_counter, next_elem;
	int rank;
	
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	//printf("[%d] -> starting communication...\n", rank);
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_north > -1) {
		next_elem = 0;
		// pack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(nz_counter = 0; nz_counter < nz + 2; nz_counter++) {
				for(nx_counter = 0; nx_counter < nx + 2; nx_counter++) {
					north_send[next_elem++] = result[nv_counter][nz_counter][ny][nx_counter];
				}
			}
		}
		
		MPI_Isend(north_send, north_size, SCALAR_MPI_TYPE, pe_north, NORTH_SEND_TAG, 
			MPI_COMM_WORLD, &requests[req_counter++]);
		MPI_Irecv(north_recv, north_size, SCALAR_MPI_TYPE, pe_north, SOUTH_SEND_TAG,
			MPI_COMM_WORLD, &requests[req_counter++]);
			
		next_elem = 0;
		
		// unpack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(nz_counter = 0; nz_counter < nz + 2; nz_counter++) {
				for(nx_counter = 0; nx_counter < nx + 2; nx_counter++) {
					result[nv_counter][nz_counter][0][nx_counter] = north_recv[next_elem++];
				}
			}
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_south > -1) {
		next_elem = 0;
		// pack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(nz_counter = 0; nz_counter < nz + 2; nz_counter++) {
				for(nx_counter = 0; nx_counter < nx + 2; nx_counter++) {
					south_send[next_elem++] = result[nv_counter][nz_counter][1][nx_counter];
				}
			}
		}
		
		MPI_Isend(south_send, south_size, SCALAR_MPI_TYPE, pe_south, SOUTH_SEND_TAG, 
			MPI_COMM_WORLD, &requests[req_counter++]);
		MPI_Irecv(south_recv, south_size, SCALAR_MPI_TYPE, pe_south, NORTH_SEND_TAG,
			MPI_COMM_WORLD, &requests[req_counter++]);
			
		next_elem = 0;
		
		// unpack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(nz_counter = 0; nz_counter < nz + 2; nz_counter++) {
				for(nx_counter = 0; nx_counter < nx + 2; nx_counter++) {
					result[nv_counter][nz_counter][ny+1][nx_counter] = south_recv[next_elem++];
				}
			}
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_east > -1) {
		next_elem = 0;
		// pack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(nz_counter = 0; nz_counter < nz + 2; nz_counter++) {
				for(ny_counter = 0; ny_counter < ny + 2; ny_counter++) {
					east_send[next_elem++] = result[nv_counter][nz_counter][ny_counter][nx];
				}
			}
		}
		
		MPI_Isend(east_send, east_size, SCALAR_MPI_TYPE, pe_east, EAST_SEND_TAG, 
			MPI_COMM_WORLD, &requests[req_counter++]);
		MPI_Irecv(east_recv, east_size, SCALAR_MPI_TYPE, pe_east, WEST_SEND_TAG,
			MPI_COMM_WORLD, &requests[req_counter++]);
			
		next_elem = 0;
		
		// unpack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(nz_counter = 0; nz_counter < nz + 2; nz_counter++) {
				for(ny_counter = 0; ny_counter < ny + 2; ny_counter++) {
					result[nv_counter][nz_counter][ny_counter][0] = east_recv[next_elem++];
				}
			}
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_west > -1) {
		next_elem = 0;
		// pack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(nz_counter = 0; nz_counter < nz + 2; nz_counter++) {
				for(ny_counter = 0; ny_counter < ny + 2; ny_counter++) {
					west_send[next_elem++] = result[nv_counter][nz_counter][ny_counter][1];
				}
			}
		}
		
		MPI_Isend(west_send, west_size, SCALAR_MPI_TYPE, pe_west, WEST_SEND_TAG, 
			MPI_COMM_WORLD, &requests[req_counter++]);
		MPI_Irecv(west_recv, west_size, SCALAR_MPI_TYPE, pe_west, EAST_SEND_TAG,
			MPI_COMM_WORLD, &requests[req_counter++]);
			
		next_elem = 0;
		
		// unpack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(nz_counter = 0; nz_counter < nz + 2; nz_counter++) {
				for(ny_counter = 0; ny_counter < ny + 2; ny_counter++) {
					result[nv_counter][nz_counter][ny_counter][nx+1] = west_recv[next_elem++];
				}
			}
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_front > -1) {
		next_elem = 0;
		// pack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(ny_counter = 0; ny_counter < ny + 2; ny_counter++) {
				for(nx_counter = 0; nx_counter < nx + 2; nx_counter++) {
					front_send[next_elem++] = result[nv_counter][1][ny_counter][nx_counter];
				}
			}
		}
		
		MPI_Isend(front_send, front_size, SCALAR_MPI_TYPE, pe_front, FRONT_SEND_TAG, 
			MPI_COMM_WORLD, &requests[req_counter++]);
		MPI_Irecv(front_recv, front_size, SCALAR_MPI_TYPE, pe_front, BACK_SEND_TAG,
			MPI_COMM_WORLD, &requests[req_counter++]);
			
		next_elem = 0;
		
		// unpack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(ny_counter = 0; ny_counter < ny + 2; ny_counter++) {
				for(nx_counter = 0; nx_counter < nx + 2; nx_counter++) {
					result[nv_counter][nz+1][ny_counter][nx_counter] = front_recv[next_elem++];
				}
			}
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////
	
	if(pe_back > -1) {
		next_elem = 0;
		// pack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(ny_counter = 0; ny_counter < ny + 2; ny_counter++) {
				for(nx_counter = 0; nx_counter < nx + 2; nx_counter++) {
					back_send[next_elem++] = result[nv_counter][nz][ny_counter][nx_counter];
				}
			}
		}
		
		MPI_Isend(back_send, back_size, SCALAR_MPI_TYPE, pe_back, BACK_SEND_TAG, 
			MPI_COMM_WORLD, &requests[req_counter++]);
		MPI_Irecv(back_recv, back_size, SCALAR_MPI_TYPE, pe_back, FRONT_SEND_TAG,
			MPI_COMM_WORLD, &requests[req_counter++]);
			
		next_elem = 0;
		
		// unpack
		for(nv_counter = 0; nv_counter < num_vars; nv_counter++) {
			for(ny_counter = 0; ny_counter < ny + 2; ny_counter++) {
				for(nx_counter = 0; nx_counter < nx + 2; nx_counter++) {
					result[nv_counter][0][ny_counter][nx_counter] = back_recv[next_elem++];
				}
			}
		}
	}
	
	
	///////////////////////////////////////////////////////////////////////////////////
	MPI_Status status[req_counter];
	MPI_Waitall(req_counter, requests, status);
	///////////////////////////////////////////////////////////////////////////////////
}